/* gtkcomboboxtext.c - GtkComboBoxText widget implementation
 *
 * Authors:               Tim Janik
 *                        Kristian Rietveld <kristian@planet.nl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkcomboboxtext.h"

#include <string.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkobject.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkbindings.h>

enum {
	SELECT,
	AUTOCOMPLETED,
	LAST_SIGNAL
};

static void      gtk_combo_box_text_destroy           (GtkObject *object);

static void      gtk_combo_box_text_class_init        (GtkComboBoxTextClass *class);
static void      gtk_combo_box_text_init              (GtkComboBoxText *cboxtext);

static GSList   *gtk_combo_box_text_build_item_list   (GtkTreeModel *model);
static void      gtk_combo_box_text_add_to_item_list  (GSList       **slist,
						       GtkTreeModel  *model,
						       GtkTreeIter   *node);
static void      gtk_combo_box_text_entry_changed     (GtkComboBoxText *cboxtext);
static gchar    *gtk_combo_box_text_node_get_value    (GtkTreeModel *model,
						       GtkTreeIter  *node);
static void      gtk_combo_box_text_node_new          (GtkComboBoxText *cboxtext,
						       const gchar     *string);
static void     gtk_combo_box_text_selection_changed  (GtkTreeView *tree_view,
						       GtkComboBoxText *cboxtext);

static gint                  gtk_combo_box_text_signals[LAST_SIGNAL] = {0};
static GtkComboBoxClass     *gtk_combo_box_text_parent_class = NULL;

/* object class functions */

static void
gtk_combo_box_text_destroy (GtkObject *object)
{
  GtkComboBoxText  *cboxtext;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (object));
  
  cboxtext = GTK_COMBO_BOX_TEXT (object);
  
  cboxtext->entry = NULL;
  cboxtext->scrolled_window = NULL;
  
  if (cboxtext->tree_view)
    gtk_widget_unref (cboxtext->tree_view);
  cboxtext->tree_view = NULL;
  
  if (cboxtext->cell)
    gtk_object_unref (GTK_OBJECT (cboxtext->cell));
  cboxtext->cell = NULL;
  
  if (cboxtext->model)
    gtk_object_unref (GTK_OBJECT (cboxtext->model));
  cboxtext->model = NULL;
  
  if (cboxtext->column)
    gtk_object_unref (GTK_OBJECT (cboxtext->column));
  cboxtext->column = NULL;
  
  GTK_OBJECT_CLASS (gtk_combo_box_text_parent_class)->destroy (object);
}

/* class functions */

static void
gtk_combo_box_text_class_init (GtkComboBoxTextClass *class)
{
  GtkObjectClass    *object_class;
  GtkWidgetClass    *widget_class;
  GtkContainerClass *container_class;
  
  g_return_if_fail (class != NULL);
  
  gtk_combo_box_text_parent_class = g_type_class_peek_parent (class);
  object_class = (GtkObjectClass *)class;
  widget_class = (GtkWidgetClass *)class;
  container_class = (GtkContainerClass *)class;
  
  object_class->destroy = gtk_combo_box_text_destroy;
  
  gtk_combo_box_text_signals[SELECT] =
    gtk_signal_new 
    ("select",
     GTK_RUN_LAST,
     gtk_combo_box_text_get_type (),
     GTK_SIGNAL_OFFSET (GtkComboBoxTextClass, select),
     gtk_signal_default_marshaller,
     GTK_TYPE_NONE, 0);
  gtk_combo_box_text_signals[AUTOCOMPLETED] =
    gtk_signal_new 
    ("autocompleted",
     GTK_RUN_LAST,
     gtk_combo_box_text_get_type (),
     GTK_SIGNAL_OFFSET (GtkComboBoxTextClass, select),
     gtk_signal_default_marshaller,
     GTK_TYPE_NONE, 0);
  
  gtk_object_class_add_signals (object_class,
				gtk_combo_box_text_signals,
				LAST_SIGNAL);
}

static void
gtk_combo_box_text_init (GtkComboBoxText *cboxtext)
{
  GtkObject    *selection;
  
  cboxtext->selected = NULL;
  cboxtext->num_items = 0;
  
  cboxtext->entry = gtk_entry_new ();
  gtk_widget_ref (cboxtext->entry);
  gtk_signal_connect_object 
    (GTK_OBJECT (cboxtext->entry), "changed",
     GTK_SIGNAL_FUNC (gtk_combo_box_text_entry_changed),
     GTK_OBJECT (cboxtext));
  gtk_widget_show (cboxtext->entry);
  
  cboxtext->scrolled_window = 
    gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy 
    (GTK_SCROLLED_WINDOW (cboxtext->scrolled_window),
     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  
  cboxtext->model = gtk_tree_store_new_with_types (1, G_TYPE_STRING);
  cboxtext->tree_view = 
    gtk_tree_view_new_with_model (GTK_TREE_MODEL (cboxtext->model));
  gtk_widget_ref (cboxtext->tree_view);
  selection = GTK_OBJECT (gtk_tree_view_get_selection 
			  (GTK_TREE_VIEW (cboxtext->tree_view )));
  gtk_tree_selection_set_type (GTK_TREE_SELECTION (selection),
			       GTK_TREE_SELECTION_MULTI);
  gtk_signal_connect (GTK_OBJECT (selection), "selection_changed",
		      GTK_SIGNAL_FUNC(gtk_combo_box_text_selection_changed),
		      cboxtext);
  gtk_container_add (GTK_CONTAINER (cboxtext->scrolled_window),
		     cboxtext->tree_view);
  
  cboxtext->cell = gtk_cell_renderer_text_new ();
  cboxtext->column = 
    gtk_tree_view_column_new_with_attributes ("nodes", cboxtext->cell, 
					      "text", 0, 
					      NULL);
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (cboxtext->tree_view),
				     FALSE);
  
  gtk_tree_view_append_column (GTK_TREE_VIEW (cboxtext->tree_view),
			       GTK_TREE_VIEW_COLUMN (cboxtext->column));

  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (cboxtext->tree_view));
  gtk_widget_size_request (cboxtext->entry, NULL);
  gtk_widget_set_usize (cboxtext->tree_view,
			cboxtext->entry->requisition.width, 14);
  gtk_widget_size_request (GTK_WIDGET (cboxtext), NULL);

  gtk_widget_show_all (cboxtext->tree_view);
  
  gtk_combo_box_set_display (GTK_COMBO_BOX (cboxtext), cboxtext->entry);
  gtk_combo_box_set_popup (GTK_COMBO_BOX (cboxtext),
			   cboxtext->scrolled_window);
}

/* utility functions */

static GSList *
gtk_combo_box_text_build_item_list (GtkTreeModel *model)
{
  GSList        *strings = NULL;
  GtkTreeIter    iter;
  
  g_return_val_if_fail (model != NULL, NULL);
  g_return_val_if_fail (GTK_IS_TREE_MODEL (model), NULL);
  
  gtk_tree_store_get_root_iter (GTK_TREE_STORE (model), &iter);
  
  gtk_combo_box_text_add_to_item_list (&strings,
				       model,
				       &iter);
  
  while (gtk_tree_model_iter_next (GTK_TREE_MODEL (model), &iter))
    gtk_combo_box_text_add_to_item_list (&strings, model, &iter);
  
  return strings;
}

static void
gtk_combo_box_text_add_to_item_list (GSList       **slist,
				     GtkTreeModel  *model,
				     GtkTreeIter   *node)
{
  gchar    *string;
  
  string = g_strdup (gtk_combo_box_text_node_get_value (model, node));
  
  if (string)
    *slist = g_slist_append (*slist, string);
  
  return;
}

static void
gtk_combo_box_text_entry_changed (GtkComboBoxText *cboxtext)
{
  gchar *tmp = gtk_entry_get_text (GTK_ENTRY (cboxtext->entry));
  
  gtk_combo_box_text_select_item (cboxtext, tmp);
}

static gchar *
gtk_combo_box_text_node_get_value (GtkTreeModel *model,
				   GtkTreeIter  *iter)
{
  GValue   value = {0,};
  
  gtk_tree_model_get_value (GTK_TREE_MODEL (model),
			    iter, 0, &value);
  
  return g_value_get_string (&value);
}

static void
gtk_combo_box_text_node_new (GtkComboBoxText *cboxtext,
			     const gchar     *string)
{
  GValue        value = {0,};
  gchar        *str;
  GtkTreeIter   iter;
  
  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  g_return_if_fail (string != NULL);
  
  g_value_init (&value, G_TYPE_STRING);
  str = g_strdup (string);
  g_value_set_string (&value, str);
  g_free (str);
  
  gtk_tree_store_append (GTK_TREE_STORE (cboxtext->model), &iter, NULL);
  gtk_tree_store_set_cell (GTK_TREE_STORE (cboxtext->model),
			   &iter, 0, &value);
  cboxtext->num_items++;
  g_value_unset (&value);

  /* resize the tree_view to something nice */
  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (cboxtext->tree_view));
  gtk_widget_size_request (cboxtext->entry, NULL);
  gtk_widget_set_usize (cboxtext->tree_view,
			cboxtext->entry->requisition.width,
			cboxtext->num_items * 14);
  gtk_widget_size_request (GTK_WIDGET (cboxtext), NULL);

  return;
}

static void
gtk_combo_box_text_selection_changed (GtkTreeView     *tree_view,
				      GtkComboBoxText *cboxtext)
{
  GtkTreeSelection  *selection;

  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  g_return_if_fail (tree_view != NULL);
  
  selection = gtk_tree_view_get_selection 
    (GTK_TREE_VIEW(cboxtext->tree_view));
  
  if (selection)
    {
      gchar       *string;
      GValue       value = {0,};
      GtkTreeIter  iter;
  
      if (gtk_tree_selection_get_selected (GTK_TREE_SELECTION (selection),
					   /* this is a bit dangerous .. */
					   (GtkTreeModel **)&cboxtext->model,
					   &iter))
	{
	  gtk_tree_model_get_value (GTK_TREE_MODEL(cboxtext->model),
				    &iter, 0, &value);
	  string = g_strdup (g_value_get_string (&value));
	  
	  if (!string)
	    return;
	  
	  gtk_entry_set_text (GTK_ENTRY (cboxtext->entry), string);
	  if (cboxtext->selected)
	    g_free (cboxtext->selected);
	  cboxtext->selected = g_strdup (string);
	  g_free (string);
	  
	  gtk_combo_box_popup_hide (GTK_COMBO_BOX (cboxtext));	
	}
    }
}


GtkType
gtk_combo_box_text_get_type (void)
{
  static GtkType   type = 0;
  
  if (!type) {
    static const GTypeInfo info = {
      sizeof (GtkComboBoxTextClass),
      NULL, /* GBaseInitFunc */
      NULL, /* GBaseFinalizeFunc */
      (GClassInitFunc) gtk_combo_box_text_class_init,
      NULL,
      NULL, /* class data */
      sizeof (GtkComboBoxText),
      0,
      (GInstanceInitFunc) gtk_combo_box_text_init
    };
    
    type = g_type_register_static (gtk_combo_box_get_type (),
				   "GtkComboBoxText",
				   &info, 0);
  }
  
  return type;
}

/**
 * gtk_comb_box_text_new:
 *
 * Returns a newly created #GtkComboBoxText.
 *
 * Return value: a #GtkWidget
 */
GtkWidget *
gtk_combo_box_text_new (void)
{
  GtkComboBoxText  *cboxtext;

  cboxtext = gtk_type_new (gtk_combo_box_text_get_type ());
  
  return GTK_WIDGET (cboxtext);
}

/**
 * gtk_combo_box_text_add_item:
 * @cboxtext: a #GtkComboBoxText
 * @string: a string
 *
 * Adds an item (the given @string) to the list
 * of @cboxtext.
 */
void
gtk_combo_box_text_add_item (GtkComboBoxText *cboxtext,
			     const gchar     *string)
{
  GtkTreeIter iter;

  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  g_return_if_fail (string != NULL);

  if (cboxtext->num_items > 0)
    {
      /* check if this item isn't already in the list */
      gtk_tree_store_get_root_iter (GTK_TREE_STORE (cboxtext->model), &iter);
      
      if (!strcmp (string, gtk_combo_box_text_node_get_value
		   (GTK_TREE_MODEL (cboxtext->model), &iter)))
	/* item is already on the list */
	return;
      
      if (cboxtext->num_items > 1)
	{
	  if (GTK_TREE_STORE (cboxtext->model)->root 
	      != GTK_TREE_STORE (cboxtext->model)->last)
	    {
	      while (gtk_tree_model_iter_next
		     (GTK_TREE_MODEL (cboxtext->model),
		      &iter))
		if (!strcmp (string, gtk_combo_box_text_node_get_value
			     (GTK_TREE_MODEL (cboxtext->model), &iter)))
		  /* item is already on the list */
		  return;
	    }
	}
    }
  
  /* add item */
  gtk_combo_box_text_node_new (cboxtext, string);
}

/**
 * gtk_combo_box_text_get_entry:
 * @cboxtext: a #GtkComboBoxText
 *
 * Returns the entry of the @cboxtext.
 *
 * Return value: a #GtkWidget.
 */
GtkWidget *
gtk_combo_box_text_get_entry (GtkComboBoxText *cboxtext)
{
  g_return_val_if_fail (cboxtext != NULL, NULL);
  g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);

  return cboxtext->entry;
}

/**
 * gtk_combo_box_text_get_selected:
 * @cboxtext: a #GtkComboBoxText
 * 
 * Returns the value of the selected item.
 *
 * Return value: a string
 */
gchar *
gtk_combo_box_text_get_selected (GtkComboBoxText *cboxtext)
{
  g_return_val_if_fail (cboxtext != NULL, NULL);
  g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);
  
  if (cboxtext->selected)
    return g_strdup (cboxtext->selected);

  return NULL;
}

/**
 * gtk_combo_box_text_get_items:
 * @cboxtext: a #GtkComboBoxText
 *
 * Returns pointers to the items of @cboxtext.
 *
 * Return value: an array of strings
 */
gchar **
gtk_combo_box_text_get_items (GtkComboBoxText *cboxtext)
{
  gint     len;
  gint     i;
  gchar  **strings;
  GSList  *list;
  GSList  *li;
  
  g_return_val_if_fail (cboxtext != NULL, NULL);
  g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);
  
  list = gtk_combo_box_text_build_item_list 
    (GTK_TREE_MODEL(cboxtext->model));
  
  g_return_val_if_fail (list != NULL, NULL);
  
  len = g_slist_length (list);
  strings = g_new0 (gchar *, len);
  g_return_val_if_fail (strings != NULL, NULL);
  
  for (i = 0, li = list ; i < len && li != NULL ; i++, li = li->next)
    strings[i] = g_strdup (li->data);

  return strings;
}

/**
 * gtk_combo_box_text_remove_item:
 * @cboxtext: a #GtkComboBoxText
 * @string: an item to remove
 *
 * Remove the item with the value of @string from the list
 * of @cboxtext.
 */
void
gtk_combo_box_text_remove_item (GtkComboBoxText *cboxtext,
				const gchar     *string)
{
  GtkTreeIter iter;
  
  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  g_return_if_fail (string != NULL);

  if (cboxtext->num_items < 1) return;

  gtk_tree_store_get_root_iter (GTK_TREE_STORE (cboxtext->model), &iter);
  
  if (!strcmp (string, gtk_combo_box_text_node_get_value 
	       (GTK_TREE_MODEL (cboxtext->model), &iter)))
    {
      gtk_tree_store_remove (GTK_TREE_STORE (cboxtext->model), &iter);
      cboxtext->num_items--;
    }
  
  if (cboxtext->num_items < 1) return;

  while (gtk_tree_model_iter_next (GTK_TREE_MODEL (cboxtext->model),
				   &iter))
    {
      if (!strcmp (string, gtk_combo_box_text_node_get_value
		   (GTK_TREE_MODEL (cboxtext->model), &iter)))
	{
	  gtk_tree_store_remove (GTK_TREE_STORE (cboxtext->model), &iter);
	  cboxtext->num_items--;
	}
    }
}

/**
 * gtk_combo_box_text_remove_matches:
 * @cboxtext: a #GtkComboBoxText
 * @pattern: a pattern
 *
 * Removes every item from the list which matches the given
 * @pattern.
 */
void
gtk_combo_box_text_remove_matches (GtkComboBoxText *cboxtext,
				   const gchar     *pattern)
{
  GtkTreeIter      iter;
  GtkPatternSpec   pspec;
  
  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  
  if (cboxtext->num_items < 1) return;

  if (!pattern)
    pattern = "*";
  
  gtk_pattern_spec_init (&pspec, pattern);	
  
  gtk_tree_store_get_root_iter (GTK_TREE_STORE (cboxtext->model), &iter);
  
  if (gtk_pattern_match_string (&pspec,
				gtk_combo_box_text_node_get_value
				(GTK_TREE_MODEL (cboxtext->model), &iter)))
    {
      gtk_tree_store_remove (GTK_TREE_STORE (cboxtext->model), &iter);
      cboxtext->num_items--;
    }
  
  if (cboxtext->num_items < 1) return;

  while (gtk_tree_model_iter_next (GTK_TREE_MODEL (cboxtext->model),
				   &iter))
    if (gtk_pattern_match_string (&pspec,
				  gtk_combo_box_text_node_get_value
				  (GTK_TREE_MODEL (cboxtext->model), &iter)))
      {
	gtk_tree_store_remove (GTK_TREE_STORE (cboxtext->model), &iter);
	cboxtext->num_items--;
      }
  
  gtk_pattern_spec_free_segs (&pspec);
}

/**
 * gtk_combo_box_text_select_item:
 * @cboxtext: a #GtkComboBoxText
 * @string: the item to select
 *
 * Selects the item with the value @string in the list if it's
 * there.
 */
void
gtk_combo_box_text_select_item (GtkComboBoxText *cboxtext,
				const gchar     *string)
{
  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
  g_return_if_fail (string != NULL);
  
  gtk_signal_emit (GTK_OBJECT (cboxtext),
		   gtk_combo_box_text_signals[SELECT],
		   string);
}
